from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Callable, cast

import numpy as np
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from numpy.typing import NDArray

from .grids import AbstractGrid

Array = NDArray[np.float64]


try:  # pragma: no cover
    from assimulo.problem import Explicit_Problem
    from assimulo.solvers import CVode

    def _simulate(
        fn: Callable[[float, Array], Array],
        y0: Array,
        t_end: float,
        t_points: int = 0,
        t_eval: Array | None = None,
    ) -> tuple[Array, Array]:
        problem = Explicit_Problem(fn, y0)
        integrator = CVode(problem)
        integrator.verbosity = 50
        t, y = integrator.simulate(t_end, t_points, t_eval)
        return np.array(t), np.array(y)

except (ImportError, ModuleNotFoundError):  # pragma: no cover
    import scipy.integrate

    warnings.warn("Assimulo not found, disabling sundials support.")

    def _simulate(
        fn: Callable[[float, Array], Array],
        y0: Array,
        t_end: float,
        t_points: int = 0,
        t_eval: Array | None = None,
    ) -> tuple[Array, Array]:
        if t_eval is None:
            if t_points > 0:
                t = np.linspace(0, t_end, t_points)
            else:
                t = np.linspace(0, t_end, 100)
        else:
            t = t_eval
        y = scipy.integrate.odeint(fn, y0, t, tfirst=True)
        return np.array(t), np.array(y)


try:  # pragma: no cover
    # from pde_perffuncs import diffusion_nonavg  # type: ignore
    raise ImportError
except ImportError:  # pragma: no cover

    def diffusion(
        y0: Array,
        is_celltype: list[bool],
        celltype_neighbors: NDArray[np.int64],  # list[list[int]]
        alpha: float,
        n_neighbors: float,
    ) -> Array:
        dydt = np.zeros_like(y0, dtype=float)
        for cell, cell_conc in enumerate(y0):
            if not is_celltype[cell]:
                continue
            diff = 0.0
            for neighbor in celltype_neighbors[cell]:
                if neighbor == -1:
                    continue
                diff += y0[neighbor] - cell_conc
            dydt[cell] += diff / n_neighbors * alpha
        return dydt

    def diffusion_nonavg(
        y0: Array,
        is_celltype: list[bool],
        celltype_neighbors: NDArray[np.int64],  # list[list[int]]
        alpha: float,
    ) -> Array:
        dydt = np.zeros_like(y0, dtype=float)
        for cell, cell_conc in enumerate(y0):
            if not is_celltype[cell]:
                continue
            diff = 0.0
            for neighbor in celltype_neighbors[cell]:
                if neighbor == -1:
                    continue
                diff += y0[neighbor] - cell_conc
            dydt[cell] += diff * alpha
        return dydt

    def diffusion_active(
        y0: Array,
        is_celltype: list[bool],
        celltype_neighbors: NDArray[np.int64],  # list[list[int]]
        alpha: float,
    ) -> Array:
        dydt = np.zeros_like(y0, dtype=float)
        for cell, cell_conc in enumerate(y0):
            if not is_celltype[cell]:
                continue
            for neighbor in celltype_neighbors[cell]:
                if neighbor == -1:
                    continue
                diff = (y0[neighbor] - cell_conc) * alpha
                if diff < 0:
                    dydt[cell] += diff  # mesophyll cell
                    dydt[neighbor] -= diff  # vein cell
        return dydt


def photosynthesis_saturating(sugar: Array, k: float, capacity: float) -> Array:
    return k * (1 - sugar / capacity)


def photosynthesis_saturating_co2(
    sugar: Array, co2: Array, k: float, capacity: float
) -> Array:
    return k * co2 * (1 - sugar / capacity)


def outflux(sugar: Array, alpha: float) -> Array:
    return -sugar * alpha


@dataclass
class AbstractModel(ABC):
    grid: AbstractGrid

    @abstractmethod
    def _get_right_hand_side(self, t: float, y0: Array) -> Array: ...

    def simulate(
        self,
        y0: Array,
        t_end: float | None = None,
        t_points: int = 0,
        t_eval: Array | None = None,
    ) -> tuple[Array, ...]:
        if t_end is None:
            if t_eval is None:
                raise ValueError("Either `t_end` or `t_eval` needs to be supplied")
            t_end = t_eval[-1]

        self.grid.initialize()

        # Do test run for better error messages
        self._get_right_hand_side(0, y0)

        t, y = _simulate(
            self._get_right_hand_side,
            y0,
            cast(float, t_end),
            t_points,
            t_eval,
        )
        n_vars = len(self.grid.variables)
        if n_vars < 2:
            return t, y
        else:
            return t, *np.split(y, n_vars, axis=1)


@dataclass
class MesophyllModel(AbstractModel):
    ps_k: float
    ps_capacity: float
    suc_meso_to_meso: float

    def __post_init__(self) -> None:
        self.grid.add_celltype("mesophyll", 1)
        self.grid.add_variable("sucrose")

    def _create_variables(self, y: Array) -> tuple[dict[str, Array], dict[str, Array]]:
        in_vars = dict(zip(self.grid.variables, np.hsplit(y, len(self.grid.variables))))
        out_vars = {
            i: np.zeros(self.grid.gridsize, dtype=float).flatten()
            for i in self.grid.variables
        }
        return in_vars, out_vars

    def _get_influxes(
        self, in_vars: dict[str, Array], out_vars: dict[str, Array]
    ) -> None:
        mesophyll_cells = self.grid.celltype_indices["mesophyll"]
        dydt = photosynthesis_saturating(
            in_vars["sucrose"][mesophyll_cells], self.ps_k, self.ps_capacity
        )
        out_vars["sucrose"][mesophyll_cells] += dydt

    def get_influxes_by_process(self, y: Array) -> dict[str, dict[str, Array]]:
        self.grid.initialize()
        mesophyll_cells = self.grid.celltype_indices["mesophyll"]
        in_vars, out_vars = self._create_variables(y)
        return {
            "sucrose": {
                "photosynthesis": photosynthesis_saturating(
                    in_vars["sucrose"][mesophyll_cells], self.ps_k, self.ps_capacity
                )
            }
        }

    def get_influxes(self, y: Array) -> dict[str, Array]:
        self.grid.initialize()
        in_vars, out_vars = self._create_variables(y)
        self._get_influxes(in_vars, out_vars)
        return out_vars

    def _get_diffusion_processes(
        self, in_vars: dict[str, Array], out_vars: dict[str, Array]
    ) -> None:
        mesophyll_cells = self.grid._is_celltype_array["mesophyll"]
        meso_to_meso = self.grid.neighbors["mesophyll", "mesophyll"]

        out_vars["sucrose"] += diffusion_nonavg(
            y0=in_vars["sucrose"],
            is_celltype=mesophyll_cells,
            celltype_neighbors=meso_to_meso,
            alpha=self.suc_meso_to_meso,
        )

    def get_diffusion_by_process(self, y: Array) -> dict[str, dict[str, Array]]:
        self.grid.initialize()
        in_vars, _ = self._create_variables(y)
        mesophyll_cells = self.grid._is_celltype_array["mesophyll"]
        meso_to_meso = self.grid.neighbors["mesophyll", "mesophyll"]
        return {
            "sucrose": {
                "meso_to_meso": diffusion_nonavg(
                    y0=in_vars["sucrose"],
                    is_celltype=mesophyll_cells,
                    celltype_neighbors=meso_to_meso,
                    alpha=self.suc_meso_to_meso,
                ),
            }
        }

    def get_diffusion(self, y: Array) -> dict[str, Array]:
        self.grid.initialize()
        in_vars, out_vars = self._create_variables(y)
        self._get_diffusion_processes(in_vars, out_vars)
        return out_vars

    def _get_right_hand_side(self, t: float, y: Array) -> Array:
        in_vars, out_vars = self._create_variables(y)
        self._get_influxes(in_vars, out_vars)
        self._get_diffusion_processes(in_vars, out_vars)
        return np.array(list(out_vars.values())).flatten()

    def get_right_hand_side(self, y: Array) -> Array:
        self.grid.initialize()
        return self._get_right_hand_side(0, y)

    def add_mesophyll_cell(self, coordinates: tuple[int, ...]) -> None:
        self.grid.add_cell(coordinates=coordinates, celltype="mesophyll")

    def _plot_all_celltypes(
        self, ax: Axes, normalized_concentrations: Array, alpha: float | None
    ) -> None:
        self._plot_mesophyll_cells(
            ax=ax,
            normalized_concentrations=normalized_concentrations,
            alpha=alpha,
        )

    def _plot_mesophyll_cells(
        self, ax: Axes, normalized_concentrations: Array, alpha: float | None
    ) -> None:
        for cr in zip(*np.where(self.grid.cells == 1)):
            if alpha is None:
                al = max(normalized_concentrations[cr], 0.05)  # type: ignore
            else:
                al = alpha
            self.grid.plot_cell(
                cr,
                facecolor=(0.00, 0.50, 0.00, al),
                edgecolor=(0, 0, 0, 1),
                ax=ax,
            )

    def plot_cells(
        self,
        ax: Axes,
        concentrations: Array | None = None,
        min_conc: float = 0.0,
        max_conc: float | None = None,
        alpha: float | None = None,
    ) -> Axes:
        if concentrations is None:
            concentrations = np.ones(self.grid.gridsize)
            min_conc = 0.0
            max_conc = 1.0

        if max_conc is None:
            max_conc = cast(float, np.max(concentrations))
        if min_conc == max_conc:
            max_conc += 0.1
        self._plot_all_celltypes(
            ax=ax,
            normalized_concentrations=(
                (concentrations - min_conc) / (max_conc - min_conc)
            ).reshape(self.grid.gridsize),
            alpha=alpha,
        )

        return ax

    def plot(
        self,
        concentrations: Array | None = None,
        min_conc: float = 0.0,
        max_conc: float | None = None,
        figsize: tuple[int, int] = (10, 10),
        ax: Axes | None = None,
        annotate: bool = False,
        alpha: float | None = None,
    ) -> tuple[Figure, Axes]:
        fig, ax = self.grid.plot_axes(figsize=figsize, ax=ax)
        ax = self.plot_cells(ax, concentrations, min_conc, max_conc, alpha)
        if annotate:
            if concentrations is None:
                raise ValueError("Supply concentrations when using annotate")
            self.grid.plot_cell_concentration(ax, concentrations)
        return fig, ax


@dataclass(kw_only=True)
class VeinModel(MesophyllModel):
    vein_base_coordinates: tuple[int, ...]
    suc_vein_to_vein: float
    suc_meso_to_vein: float
    _vein_base_idx: int = 0

    def __post_init__(self) -> None:
        super().__post_init__()

        self.grid.add_celltype("vein", 2)
        self._vein_base_idx = self.grid.get_index_of_coordinates(
            self.vein_base_coordinates
        )
        self.add_vein_cell(self.vein_base_coordinates)

    def _get_diffusion_processes(
        self, in_vars: dict[str, Array], out_vars: dict[str, Array]
    ) -> None:
        mesophyll_cells = self.grid._is_celltype_array["mesophyll"]
        vein_cells = self.grid._is_celltype_array["vein"]

        meso_to_meso = self.grid.neighbors["mesophyll", "mesophyll"]
        meso_to_vein = self.grid.neighbors["mesophyll", "vein"]
        vein_to_vein = self.grid.neighbors["vein", "vein"]

        out_vars["sucrose"] += diffusion_nonavg(
            y0=in_vars["sucrose"],
            is_celltype=mesophyll_cells,
            celltype_neighbors=meso_to_meso,
            alpha=self.suc_meso_to_meso,
        )
        out_vars["sucrose"] += diffusion_active(
            y0=in_vars["sucrose"],
            is_celltype=mesophyll_cells,
            celltype_neighbors=meso_to_vein,
            alpha=self.suc_meso_to_vein,
        )
        out_vars["sucrose"] += diffusion_nonavg(
            y0=in_vars["sucrose"],
            is_celltype=vein_cells,
            celltype_neighbors=vein_to_vein,
            alpha=self.suc_vein_to_vein,
        )

    def get_diffusion_by_process(self, y: Array) -> dict[str, dict[str, Array]]:
        self.grid.initialize()
        in_vars, _ = self._create_variables(y)
        mesophyll_cells = self.grid._is_celltype_array["mesophyll"]
        vein_cells = self.grid._is_celltype_array["vein"]

        meso_to_meso = self.grid.neighbors["mesophyll", "mesophyll"]
        meso_to_vein = self.grid.neighbors["mesophyll", "vein"]
        vein_to_vein = self.grid.neighbors["vein", "vein"]
        return {
            "sucrose": {
                "meso_to_meso": diffusion_nonavg(
                    y0=in_vars["sucrose"],
                    is_celltype=mesophyll_cells,
                    celltype_neighbors=meso_to_meso,
                    alpha=self.suc_meso_to_meso,
                ),
                "meso_to_vein": diffusion_active(
                    y0=in_vars["sucrose"],
                    is_celltype=mesophyll_cells,
                    celltype_neighbors=meso_to_vein,
                    alpha=self.suc_meso_to_vein,
                ),
                "vein_to_vein": diffusion_nonavg(
                    y0=in_vars["sucrose"],
                    is_celltype=vein_cells,
                    celltype_neighbors=vein_to_vein,
                    alpha=self.suc_vein_to_vein,
                ),
            }
        }

    def _get_outfluxes(
        self, in_vars: dict[str, Array], out_vars: dict[str, Array]
    ) -> None:
        out_vars["sucrose"][self._vein_base_idx] += outflux(
            in_vars["sucrose"][self._vein_base_idx],
            self.suc_vein_to_vein,
        )

    def get_outfluxes(self, y: Array) -> dict[str, Array]:
        self.grid.initialize()
        in_vars, out_vars = self._create_variables(y)
        self._get_outfluxes(in_vars, out_vars)
        return out_vars

    def _get_right_hand_side(self, t: float, y: Array) -> Array:
        in_vars, out_vars = self._create_variables(y)
        self._get_influxes(in_vars, out_vars)
        self._get_diffusion_processes(in_vars, out_vars)
        self._get_outfluxes(in_vars, out_vars)
        return np.array(list(out_vars.values())).flatten()

    def add_vein_cell(self, coordinates: tuple[int, ...]) -> None:
        self.grid.add_cell(coordinates=coordinates, celltype="vein")

    def get_vein_outflux(self, y: Array) -> Array:
        self.grid.initialize()
        """Get vein outflux for a single concentration array"""

        if y.ndim > 1:
            return -outflux(
                y[:, self._vein_base_idx],
                self.suc_vein_to_vein,
            )
        return -outflux(
            y[self._vein_base_idx],
            self.suc_vein_to_vein,
        )

    def _plot_vein_cells(
        self, ax: Axes, normalized_concentrations: Array, alpha: float | None
    ) -> None:
        for cr in zip(*np.where(self.grid.cells == 2)):
            if alpha is None:
                al = max(normalized_concentrations[cr], 0.05)  # type: ignore
            else:
                al = alpha
            self.grid.plot_cell(
                cr,
                facecolor=(0.30, 0.15, 0.03, al),
                edgecolor=(0, 0, 0, 1),
                ax=ax,
            )

    def _plot_all_celltypes(
        self, ax: Axes, normalized_concentrations: Array, alpha: float | None
    ) -> None:
        self._plot_mesophyll_cells(
            ax=ax,
            normalized_concentrations=normalized_concentrations,
            alpha=alpha,
        )
        self._plot_vein_cells(
            ax=ax,
            normalized_concentrations=normalized_concentrations,
            alpha=alpha,
        )


@dataclass(kw_only=True)
class StomataModel(VeinModel):
    co2_k: float
    co2_capacity: float
    co2_stoma_to_meso: float
    co2_meso_to_meso: float

    def __post_init__(self) -> None:
        super().__post_init__()

        self.grid.add_celltype("stoma", 3)
        self.grid.add_variable("co2")

    def _get_influxes(
        self, in_vars: dict[str, Array], out_vars: dict[str, Array]
    ) -> None:
        mesophyll_cells = self.grid.celltype_indices["mesophyll"]
        stoma_cells = self.grid.celltype_indices["stoma"]

        # Sucrose processes
        dydt = photosynthesis_saturating_co2(
            sugar=in_vars["sucrose"][mesophyll_cells],
            co2=in_vars["co2"][mesophyll_cells],
            k=self.ps_k,
            capacity=self.ps_capacity,
        )
        out_vars["sucrose"][mesophyll_cells] += dydt
        out_vars["co2"][mesophyll_cells] -= dydt

        # CO2 processes
        out_vars["co2"][stoma_cells] += photosynthesis_saturating(
            sugar=in_vars["co2"][stoma_cells],
            k=self.co2_k,
            capacity=self.co2_capacity,
        )

    def get_influxes_by_process(self, y: Array) -> dict[str, dict[str, Array]]:
        self.grid.initialize()
        mesophyll_cells = self.grid.celltype_indices["mesophyll"]
        in_vars, out_vars = self._create_variables(y)

        mesophyll_cells = self.grid.celltype_indices["mesophyll"]
        stoma_cells = self.grid.celltype_indices["stoma"]

        return {
            "sucrose": {
                "photosynthesis": photosynthesis_saturating_co2(
                    sugar=in_vars["sucrose"][mesophyll_cells],
                    co2=in_vars["co2"][mesophyll_cells],
                    k=self.ps_k,
                    capacity=self.ps_capacity,
                ),
            },
            "co2": {
                "stomata": photosynthesis_saturating(
                    sugar=in_vars["co2"][stoma_cells],
                    k=self.co2_k,
                    capacity=self.co2_capacity,
                ),
            },
        }

    def _get_diffusion_processes(
        self, in_vars: dict[str, Array], out_vars: dict[str, Array]
    ) -> None:
        mesophyll_cells = self.grid._is_celltype_array["mesophyll"]
        vein_cells = self.grid._is_celltype_array["vein"]
        stoma_cells = self.grid._is_celltype_array["stoma"]

        meso_to_meso = self.grid.neighbors["mesophyll", "mesophyll"]
        meso_to_vein = self.grid.neighbors["mesophyll", "vein"]
        vein_to_vein = self.grid.neighbors["vein", "vein"]
        stoma_to_meso = self.grid.neighbors["stoma", "mesophyll"]

        out_vars["sucrose"] += diffusion_nonavg(
            y0=in_vars["sucrose"],
            is_celltype=mesophyll_cells,
            celltype_neighbors=meso_to_meso,
            alpha=self.suc_meso_to_meso,
        )
        out_vars["sucrose"] += diffusion_active(
            y0=in_vars["sucrose"],
            is_celltype=mesophyll_cells,
            celltype_neighbors=meso_to_vein,
            alpha=self.suc_meso_to_vein,
        )
        out_vars["sucrose"] += diffusion_nonavg(
            y0=in_vars["sucrose"],
            is_celltype=vein_cells,
            celltype_neighbors=vein_to_vein,
            alpha=self.suc_vein_to_vein,
        )
        out_vars["co2"] += diffusion_active(
            y0=in_vars["co2"],
            is_celltype=stoma_cells,
            celltype_neighbors=stoma_to_meso,
            alpha=self.co2_stoma_to_meso,
        )
        out_vars["co2"] += diffusion_nonavg(
            y0=in_vars["co2"],
            is_celltype=mesophyll_cells,
            celltype_neighbors=meso_to_meso,
            alpha=self.co2_meso_to_meso,
        )

    def get_diffusion_by_process(self, y: Array) -> dict[str, dict[str, Array]]:
        self.grid.initialize()
        in_vars, _ = self._create_variables(y)

        mesophyll_cells = self.grid._is_celltype_array["mesophyll"]
        vein_cells = self.grid._is_celltype_array["vein"]
        stoma_cells = self.grid._is_celltype_array["stoma"]

        meso_to_meso = self.grid.neighbors["mesophyll", "mesophyll"]
        meso_to_vein = self.grid.neighbors["mesophyll", "vein"]
        vein_to_vein = self.grid.neighbors["vein", "vein"]
        stoma_to_meso = self.grid.neighbors["stoma", "mesophyll"]
        return {
            "sucrose": {
                "meso_to_meso": diffusion_nonavg(
                    y0=in_vars["sucrose"],
                    is_celltype=mesophyll_cells,
                    celltype_neighbors=meso_to_meso,
                    alpha=self.suc_meso_to_meso,
                ),
                "meso_to_vein": diffusion_active(
                    y0=in_vars["sucrose"],
                    is_celltype=mesophyll_cells,
                    celltype_neighbors=meso_to_vein,
                    alpha=self.suc_meso_to_vein,
                ),
                "vein_to_vein": diffusion_nonavg(
                    y0=in_vars["sucrose"],
                    is_celltype=vein_cells,
                    celltype_neighbors=vein_to_vein,
                    alpha=self.suc_vein_to_vein,
                ),
            },
            "co2": {
                "stoma_to_meso": diffusion_active(
                    y0=in_vars["co2"],
                    is_celltype=stoma_cells,
                    celltype_neighbors=stoma_to_meso,
                    alpha=self.co2_stoma_to_meso,
                ),
                "co2_meso_to_meso": diffusion_nonavg(
                    y0=in_vars["co2"],
                    is_celltype=mesophyll_cells,
                    celltype_neighbors=meso_to_meso,
                    alpha=self.co2_meso_to_meso,
                ),
            },
        }

    def add_stoma_cell(self, coordinates: tuple[int, ...]) -> None:
        self.grid.add_cell(coordinates=coordinates, celltype="stoma")

    def _plot_stomata_cells(
        self, ax: Axes, normalized_concentrations: Array, alpha: float | None
    ) -> None:
        for cr in zip(*np.where(self.grid.cells == 3)):
            if alpha is None:
                al = max(normalized_concentrations[cr], 0.05)  # type: ignore
            else:
                al = alpha
            self.grid.plot_cell(
                cr,
                facecolor=(0.80, 0.37, 0.04, al),
                edgecolor=(0, 0, 0, 1),
                ax=ax,
            )

    def _plot_all_celltypes(
        self, ax: Axes, normalized_concentrations: Array, alpha: float | None
    ) -> None:
        self._plot_mesophyll_cells(
            ax=ax,
            normalized_concentrations=normalized_concentrations,
            alpha=alpha,
        )
        self._plot_vein_cells(
            ax=ax,
            normalized_concentrations=normalized_concentrations,
            alpha=alpha,
        )
        self._plot_stomata_cells(
            ax=ax,
            normalized_concentrations=normalized_concentrations,
            alpha=alpha,
        )



from __future__ import annotations

import itertools
import math
from abc import ABC, abstractmethod
from typing import Iterable, cast

import matplotlib.patches as patches
import numpy as np
from matplotlib.axes import Axes
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from typing_extensions import Self

PlotCoordinates = tuple[float, ...]


class _RegularPolygon(ABC):
    def __init__(self, *coordinates: int) -> None:
        self.grid_coordinates = coordinates

    def __str__(self) -> str:
        return f"{self.grid_coordinates}"

    def __repr__(self) -> str:
        return f"{self.grid_coordinates}"

    def __eq__(self, other: object) -> bool:
        assert isinstance(other, _RegularPolygon)
        return self.grid_coordinates == other.grid_coordinates

    # def __add__(self, other: Self) -> Self:
    #     return self.__class__(
    #         *(i + j for i, j in zip(self.grid_coordinates, other.grid_coordinates))
    #     )

    # def __sub__(self, other: Self) -> Self:
    #     return self.__class__(
    #         *(i - j for i, j in zip(self.grid_coordinates, other.grid_coordinates))
    #     )

    # def __mul__(self, other: Self) -> Self:
    #     return self.__class__(
    #         *(i * j for i, j in zip(self.grid_coordinates, other.grid_coordinates))
    #     )

    # def __pow__(self, other: Self) -> Self:
    #     return self.__class__(
    #         *(i**j for i, j in zip(self.grid_coordinates, other.grid_coordinates))
    #     )

    # def __floordiv__(self, other: Self) -> Self:
    #     return self.__class__(
    #         *(i // j for i, j in zip(self.grid_coordinates, other.grid_coordinates))
    #     )

    # def __mod__(self, other: Self) -> Self:
    #     return self.__class__(
    #         *(i % j for i, j in zip(self.grid_coordinates, other.grid_coordinates))
    #     )

    # def __iter__(self) -> Iterable[float]:
    #     yield from self.grid_coordinates

    @abstractmethod
    def neighbors(self) -> list[Self]:
        ...

    @abstractmethod
    def to_plot_coordinates(self) -> PlotCoordinates:
        ...

    @abstractmethod
    def plot(
        self,
        ax: Axes,
        facecolor: str | tuple[float, float, float, float],
        edgecolor: str | tuple[float, float, float, float],
    ) -> None:
        ...

    @abstractmethod
    def distance(self, other: Self) -> float:
        ...


###############################################################################
# ROD COORDINATES (1D)
###############################################################################


class Rod(_RegularPolygon):
    def __init__(self, x: int) -> None:
        super().__init__(x)
        self.L = 1

    def neighbors(self) -> list[Rod]:
        x = self.grid_coordinates[0]
        return [Rod(x - 1), Rod(x + 1)]

    def to_plot_coordinates(self) -> PlotCoordinates:
        return self.grid_coordinates[0], 0

    def plot(
        self,
        ax: Axes,
        facecolor: str | tuple[float, float, float, float],
        edgecolor: str | tuple[float, float, float, float],
    ) -> None:
        ax.add_patch(
            patches.RegularPolygon(
                self.to_plot_coordinates(),  # type: ignore
                numVertices=4,
                radius=math.sin(math.pi / 4) * self.L,
                facecolor=facecolor,
                edgecolor=edgecolor,
                orientation=math.radians(45),
            )
        )

    def distance(self, other: Self) -> float:
        return abs(self.grid_coordinates[0] - other.grid_coordinates[0])


###############################################################################
# TRIANGLE COORDINATES
###############################################################################


class Triangle(_RegularPolygon):
    def __init__(self, col: int, row: int) -> None:
        super().__init__(col, row)
        self.L = 1  # Side length
        self.d = self.L / math.sqrt(3)  # Radius
        self.h = self.L / 2 * math.sqrt(3)  # Height
        self.a = self.h - self.d  # Distance from middle to base
        self.num_vertices = 3
        self.orientation = math.radians(180 * (col % 2 + row % 2))
        self.x_offset = 0.5 * self.L
        self.y_offset = self.a

    def plot(
        self,
        ax: Axes,
        facecolor: str | tuple[float, float, float, float] = "C1",
        edgecolor: str | tuple[float, float, float, float] = (0, 0, 0, 1),
    ) -> None:
        ax.add_patch(
            patches.RegularPolygon(
                self.to_plot_coordinates(),  # type: ignore
                numVertices=self.num_vertices,
                radius=self.d,
                facecolor=facecolor,
                edgecolor=edgecolor,
                orientation=self.orientation,
            )
        )

    def to_plot_coordinates(self) -> PlotCoordinates:
        c, r = self.grid_coordinates
        x = c / 2
        # Only add a if either r or c is odd
        y = r * self.h + (self.a * (cast(int, r) % 2 ^ cast(int, c) % 2))
        return x, y

    def neighbors(self) -> list[Triangle]:
        c, r = self.grid_coordinates
        if c % 2 == r % 2:
            return [Triangle(c + 1, r), Triangle(c, r - 1), Triangle(c - 1, r)]
        return [Triangle(c, r + 1), Triangle(c + 1, r), Triangle(c - 1, r)]

    def distance(self, other: Triangle) -> float:
        """Manhattan distance"""
        c1, r1 = self.grid_coordinates
        c2, r2 = other.grid_coordinates
        return (
            abs(r1 - r2)
            + abs(r1 + (-c1 - r1) / 2 - r2 - (-c2 - r2) / 2)
            + abs((c1 + r1) / 2 - (c2 + r2) / 2)
        )


###############################################################################
# SQUARE COORDINATES
###############################################################################


class Square(_RegularPolygon):
    def __init__(self, x: int, y: int) -> None:
        super().__init__(x, y)
        self.L = 1  # side length
        self.r = math.sin(math.pi / 4) * self.L  # Radius
        self.num_vertices = 4
        self.orientation = math.radians(45)
        self.x_offset = 0.5 * self.L
        self.y_offset = 0.5 * self.L

    def plot(
        self,
        ax: Axes,
        facecolor: str | tuple[float, float, float, float] = "C1",
        edgecolor: str | tuple[float, float, float, float] = (0, 0, 0, 1),
    ) -> None:
        ax.add_patch(
            patches.RegularPolygon(
                self.to_plot_coordinates(),  # type: ignore
                numVertices=self.num_vertices,
                radius=self.r,
                facecolor=facecolor,
                edgecolor=edgecolor,
                orientation=self.orientation,
            )
        )

    def to_oddq(self) -> Oddq:
        x, y = self.grid_coordinates
        c = 2 / 3 * x
        r = y / math.sqrt(3) - 0.5 * (c % 2)
        return Oddq(int(c), int(r))

    def to_triangle(self) -> Triangle:
        x, y = self.grid_coordinates
        t = Triangle(0, 0)  # Dummy triangle to get height and a
        c = (x - 0.5) * 2
        r = y - t.a
        # Remainder will be 0 or 0.333.., so I can just round it down
        r = round(r / t.h)
        return Triangle(int(c), int(r))

    def to_plot_coordinates(self) -> PlotCoordinates:
        return self.grid_coordinates

    def neighbors(
        self,
        directions: Iterable[int] = (0, 1, 2, 3),
    ) -> list[Square]:
        x, y = self.grid_coordinates
        surroundings = ((-1, 0), (0, 1), (1, 0), (0, -1))
        result = []
        for direction in directions:
            i, j = surroundings[direction]
            result.append(Square(x + i, y + j))
        return result

    def distance(self, other: Self) -> int:
        """Manhattan distance"""
        x1, y1 = self.grid_coordinates
        x2, y2 = other.grid_coordinates
        return abs(x1 - x2) + abs(y1 - y2)  # type: ignore


###############################################################################
# HEXAGON COORDINATES
###############################################################################


class Oddq(_RegularPolygon):
    def __init__(self, col: int, row: int) -> None:
        super().__init__(col, row)
        self.L = 1
        self.radius = 1
        self.num_vertices = 6
        self.orientation = math.radians(30)

    def plot(
        self,
        ax: Axes,
        facecolor: str | tuple[float, float, float, float] = "C1",
        edgecolor: str | tuple[float, float, float, float] = (0, 0, 0, 1),
    ) -> None:
        x, y = self.to_plot_coordinates()
        ax.add_patch(
            patches.RegularPolygon(
                (x, y),
                numVertices=self.num_vertices,
                radius=self.radius,
                facecolor=facecolor,
                edgecolor=edgecolor,
                orientation=self.orientation,
            )
        )

    def to_plot_coordinates(self) -> PlotCoordinates:
        c, r = self.grid_coordinates
        x = 3 / 2 * c
        y = math.sqrt(3) * (r + 0.5 * (c % 2))  # If column is even, shift row
        return float(x), float(y)

    def to_cube(self) -> Cube:
        c, r = self.grid_coordinates
        x = c
        z = r - (c - c % 2) / 2
        y = -x - z
        return Cube(int(x), int(y), int(z))

    def to_axial(self) -> Axial:
        c, r = self.grid_coordinates
        return Axial(int(c), int(r - (c - c % 2) / 2))

    def to_doubleheight(self) -> Doubleheight:
        c, r = self.grid_coordinates
        return Doubleheight(int(c), int(r * 2 + (c % 2)))

    def neighbors(self, directions: Iterable[int] = (0, 1, 2, 3, 4, 5)) -> list[Oddq]:
        """Neighbors of a cell in odd-q coordinates.
        Coordinates start at the upper right and go clockwise
        """
        c, r = self.grid_coordinates
        parity = cast(int, c % 2)
        surroundings = (
            ((1, 0), (1, -1), (0, -1), (-1, -1), (-1, 0), (0, 1)),
            ((1, 1), (1, 0), (0, -1), (-1, 0), (-1, 1), (0, 1)),
        )

        result = []
        for direction in directions:
            i, j = surroundings[parity][direction]
            result.append(Oddq(c + i, r + j))
        return result

    def distance(self, other: Oddq) -> float:
        """Manhattan distance"""
        return self.to_cube().distance(other.to_cube())


class Cube(_RegularPolygon):
    def __init__(self, x: int, y: int, z: int) -> None:
        super().__init__(x, y, z)

    def to_oddq(self) -> Oddq:
        x, _, z = self.grid_coordinates
        c = x
        r = z + (x - x % 2) / 2
        return Oddq(int(c), int(r))

    def to_axial(self) -> Axial:
        x, _, z = self.grid_coordinates
        return Axial(int(x), int(z))

    def to_doubleheight(self) -> Doubleheight:
        x, _, z = self.grid_coordinates
        c = x
        r = 2 * z + x
        return Doubleheight(c, r)

    def neighbors(self, directions: Iterable[int] = (0, 1, 2, 3, 4, 5)) -> list[Cube]:
        """Neighbors of a cell in cube coordinates.
        Coordinates start at the upper right and go clockwise
        """
        x, y, z = self.grid_coordinates
        surroundings = (
            (1, -1, 0),
            (1, 0, -1),
            (0, 1, -1),
            (-1, 1, 0),
            (-1, 0, 1),
            (0, -1, 1),
        )
        result = []
        for direction in directions:
            i, j, k = surroundings[direction]
            result.append(Cube(x + i, y + j, z + k))
        return result

    def distance(self, cube2: Cube) -> float:
        """Distance between two cells c1 and c2 in cube coordinates.
        In a cube grid, Manhattan distances are abs(dx) + abs(dy) + abs(dz).
        The distance on a hex grid is half that:
        d = (abs(x1 - x2) + abs(y1 - y1) + abs(z1 - z2)) / 2
        An equivalent way to write this is by noting that one of the three coordinates
        must be the sum of the other two, then picking that one as the distance:
        d = max(abs(x1 - x2), abs(y1 - y2), abs(z1 - z2))
        """
        x1, y1, z1 = self.grid_coordinates
        x2, y2, z2 = cube2.grid_coordinates
        return max(abs(x1 - x2), abs(y1 - y2), abs(z1 - z2))

    def ring(self, radius: int) -> list[Cube]:
        """Gives the coordinates of all cells surrounding the center cell with radius r
        Expects cube coordinates.
        """
        x, y, z = self.grid_coordinates
        results = []
        # Go radius * steps in direction +z (4)
        x -= radius
        z += radius

        # Go around each edge
        for direction in range(6):
            # For radius steps
            for _ in range(radius):
                c = Cube(x, y, z)
                results.append(c)
                x, y, z = c.neighbors([direction])[0].grid_coordinates
        return results

    def to_plot_coordinates(self) -> PlotCoordinates:
        return self.grid_coordinates

    def plot(
        self,
        ax: Axes,
        facecolor: str | tuple[float, float, float, float],
        edgecolor: str | tuple[float, float, float, float],
    ) -> None:
        return None  # FIXME: implement


class Axial(_RegularPolygon):
    def __init__(self, q: int, r: int) -> None:
        super().__init__(q, r)

    def to_oddq(self) -> Oddq:
        q, r = self.grid_coordinates
        return Oddq(int(q), int(r + (q - q % 2) / 2))

    def to_cube(self) -> Cube:
        q, r = self.grid_coordinates
        return Cube(int(q), int(-q - r), int(r))

    def neighbors(self, directions: Iterable[int] = (0, 1, 2, 3, 4, 5)) -> list[Axial]:
        """Neighbors of a cell in axial coordinates
        Coordinates start at the upper right and go clockwise
        """
        q, r = self.grid_coordinates
        surroundings = ((1, 0), (1, -1), (0, -1), (-1, 0), (-1, 1), (0, 1))
        result = []
        for direction in directions:
            i, j = surroundings[direction]
            result.append(Axial(q + i, r + j))
        return result

    def distance(self, other: Axial) -> float:
        """Manhattan distance"""
        return self.to_cube().distance(other.to_cube())

    def to_plot_coordinates(self) -> PlotCoordinates:
        return self.grid_coordinates

    def plot(
        self,
        ax: Axes,
        facecolor: str | tuple[float, float, float, float],
        edgecolor: str | tuple[float, float, float, float],
    ) -> None:
        return None  # FIXME: implement


class Doubleheight(_RegularPolygon):
    def __init__(self, col: int, row: int) -> None:
        super().__init__(col, row)

    def to_oddq(self) -> Oddq:
        c, r = self.grid_coordinates
        return Oddq(int(c), int((r - (c % 2)) / 2))

    def to_cube(self) -> Cube:
        c, r = self.grid_coordinates
        x = c
        z = (r - c) / 2
        y = -x - z
        return Cube(x, int(y), int(z))

    def neighbors(
        self, directions: Iterable[int] = (0, 1, 2, 3, 4, 5)
    ) -> list[Doubleheight]:
        surroundings = (
            (+1, +1),
            (+1, -1),
            (0, -2),
            (-1, -1),
            (-1, +1),
            (0, +2),
        )
        c, r = self.grid_coordinates
        result = []
        for direction in directions:
            i, j = surroundings[direction]
            result.append(Doubleheight(c + i, r + j))
        return result

    def to_plot_coordinates(self) -> PlotCoordinates:
        return self.grid_coordinates

    def distance(self, other: Self) -> float:
        """Manhattan distance"""
        return self.to_cube().distance(other.to_cube())

    def plot(
        self,
        ax: Axes,
        facecolor: str | tuple[float, float, float, float],
        edgecolor: str | tuple[float, float, float, float],
    ) -> None:
        return None  # FIXME: implement


###############################################################################
# 3D coordinates
###############################################################################


class Cube3D(_RegularPolygon):
    def __init__(self, x: int, y: int, z: int) -> None:
        super().__init__(x, y, z)

    def distance(self, other: Self) -> float:
        return 1.0  # FIXME: implement

    def to_plot_coordinates(self) -> PlotCoordinates:
        x, y, z = self.grid_coordinates
        return x + 0.5, y + 0.5, z + 0.5

    def plot(
        self,
        ax: Axes,
        facecolor: str | tuple[float, float, float, float] = "C1",
        edgecolor: str | tuple[float, float, float, float] = (0, 0, 0, 1),
    ) -> None:
        cube = [
            [[0, 1, 0], [0, 0, 0], [1, 0, 0], [1, 1, 0]],
            [[0, 0, 0], [0, 0, 1], [1, 0, 1], [1, 0, 0]],
            [[1, 0, 1], [1, 0, 0], [1, 1, 0], [1, 1, 1]],
            [[0, 0, 1], [0, 0, 0], [0, 1, 0], [0, 1, 1]],
            [[0, 1, 0], [0, 1, 1], [1, 1, 1], [1, 1, 0]],
            [[0, 1, 1], [0, 0, 1], [1, 0, 1], [1, 1, 1]],
        ]

        cube = np.array(cube) + np.array(self.grid_coordinates)

        ax.add_collection3d(  # type: ignore
            Poly3DCollection(
                cube,
                facecolors=list(itertools.repeat(facecolor, 6)),
                edgecolor=edgecolor,
            )
        )

    def neighbors(self, directions: Iterable[int] = (0, 1, 2, 3, 4, 5)) -> list[Cube3D]:
        x, y, z = self.grid_coordinates
        surroundings = (
            (-1, 0, 0),
            (0, 1, 0),
            (1, 0, 0),
            (0, -1, 0),
            (0, 0, 1),
            (0, 0, -1),
        )
        result = []
        for direction in directions:
            i, j, k = surroundings[direction]
            result.append(Cube3D(x + i, y + j, z + k))
        return result



from __future__ import annotations

__all__ = [
    "AbstractGrid",
    "HexGrid",
    "RodGrid",
    "SquareGrid",
    "SquareGrid",
    "SquareGrid",
]

import copy
import itertools
import math
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Self, Type, cast

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from numpy.typing import NDArray

from .coordinates import Cube3D, Oddq, Rod, Square, Triangle, _RegularPolygon


@dataclass
class AbstractGrid(ABC):
    """General grid model representation"""

    gridsize: tuple[int, ...]
    cells: NDArray = field(default_factory=lambda: np.empty(0))
    variables: list[str] = field(default_factory=list)
    neighbors: dict[tuple[str, str], NDArray] = field(default_factory=dict)
    celltype_indices: dict[str, NDArray] = field(default_factory=dict)
    celltypes: dict[str, int] = field(default_factory=dict)
    initialized: bool = True
    _is_celltype_array: dict[str, list[bool]] = field(default_factory=dict)
    _base_shape: Type[_RegularPolygon] = _RegularPolygon  # type: ignore
    _n_neighbors: int = field(default=0)

    def __post_init__(self) -> None:
        self.initialized = True
        if self.cells.shape == (0,):
            self.cells = np.zeros(self.gridsize, dtype="int")

    @abstractmethod
    def plot_axes(
        self,
        figsize: tuple[int, int] | None = None,
        ax: Axes | None = None,
    ) -> tuple[Figure, Axes]:
        ...

    def copy(self) -> Self:
        return copy.deepcopy(self)

    def add_celltype(self, name: str, value: int) -> None:
        self.celltypes[name] = value
        self.celltype_indices[name] = np.empty(0)
        self._is_celltype_array[name] = []
        self.initialized = False

    def add_cell(self, coordinates: tuple[int, ...], celltype: str) -> None:
        self.cells[coordinates] = self.celltypes[celltype]
        self.initialized = False

    def remove_cell(self, coordinates: tuple[int, ...]) -> None:
        self.cells[coordinates] = 0
        self.initialized = False

    def add_variable(self, variable_name: str) -> None:
        if variable_name not in self.variables:
            self.variables.append(variable_name)
            self.initialized = False

    def get_celltypes(self) -> dict[str, int]:
        return self.celltypes

    def get_index_of_coordinates(self, coordinates: tuple[int, ...]) -> Any:
        return np.ravel_multi_index(coordinates, self.gridsize)

    def get_coordinates_of_index(self, index: int) -> tuple[int, ...]:
        return np.unravel_index(index, self.gridsize)  # type: ignore

    def get_indices_of_celltype(self, celltype: str) -> list[int]:
        type_value = self.celltypes[celltype]
        return [
            self.get_index_of_coordinates(i)  # type: ignore
            for i in zip(*np.where(self.cells == type_value))
        ]

    def get_celltype_of_index(self, index: int) -> NDArray:
        return self.cells[self.get_coordinates_of_index(index)]  # type: ignore

    def get_cell_neighbors(self, idx: int) -> NDArray:
        """Fills empty positions with -1"""
        upper_borders = self.cells.shape
        neighbors = np.full(self._n_neighbors, -1)
        for idx, neighbor in enumerate(
            self._base_shape(*self.get_coordinates_of_index(idx)).neighbors()
        ):
            neighbor_coordinates = neighbor.grid_coordinates
            if any((i < 0 for i in neighbor_coordinates)):
                pass
            elif any((i >= j for i, j in zip(neighbor_coordinates, upper_borders))):
                pass
            elif self.cells[neighbor_coordinates] != 0:
                neighbors[idx] = self.get_index_of_coordinates(neighbor_coordinates)
        return neighbors

    def get_cell_neighbors_of_celltype(self, idx: int, celltype: str) -> NDArray:
        ct = self.celltypes[celltype]
        neighbors = np.full(self._n_neighbors, -1)
        for idx, i in enumerate(self.get_cell_neighbors(idx)):
            if i != -1:
                if self.cells[self.get_coordinates_of_index(i)] == ct:
                    neighbors[idx] = i
        return neighbors

    def get_all_neighbors_of_celltype(
        self, celltype: str, neighbor_celltype: str
    ) -> NDArray:
        arr = np.full((np.prod(self.gridsize), self._n_neighbors), -1)
        for idx in self.get_indices_of_celltype(celltype):
            arr[idx] = self.get_cell_neighbors_of_celltype(idx, neighbor_celltype)
        return arr

    def nearest_neighbor_of_celltype(
        self, cell: tuple[int, ...], neighbor_celltype: str
    ) -> float:
        start = self._base_shape(*cell)
        return min(
            [
                start.distance(self._base_shape(*cr))
                for cr in zip(
                    *np.where((self.cells == self.celltypes[neighbor_celltype]))
                )
            ]
        )

    def nearest_distances_of_celltypes(
        self, celltype1: str, celltype2: str
    ) -> list[float]:
        return [
            self.nearest_neighbor_of_celltype(cr, celltype2)
            for cr in zip(*np.where((self.cells == self.celltypes[celltype1])))
        ]

    def initialize(self) -> None:
        """
        All functions need boolean arrays of cells
        Diffusion and active transport further need neighbor arrays
        """
        # Generate celltype indices
        for celltype in self.celltypes:
            self.celltype_indices[celltype] = np.array(
                self.get_indices_of_celltype(celltype)
            )
            # FIXME: pyo3 doesn't like the boolean array
            self._is_celltype_array[celltype] = [
                bool(i) for i in (self.cells.flatten() == self.celltypes[celltype])
            ]

        # Generate neighbor arrays
        for in_celltype in self.celltypes:
            for out_celltype in self.celltypes:
                self.neighbors[
                    in_celltype, out_celltype
                ] = self.get_all_neighbors_of_celltype(in_celltype, out_celltype)
        self.initialized = True

    def generate_initial_values(self) -> NDArray:
        # Let's start with a simple: Everything is zero ;)
        return np.zeros((*self.gridsize, len(self.variables))).flatten()

    def plot_cell(
        self,
        coordinates: tuple[int, ...],
        ax: Axes,
        facecolor: str | tuple[float, float, float, float] = "C1",
        edgecolor: str | tuple[float, float, float, float] = (0, 0, 0, 1),
    ) -> None:
        self._base_shape(*coordinates).plot(
            ax=ax, facecolor=facecolor, edgecolor=edgecolor
        )

    def plot_grid(self, ax: Axes) -> None:
        cols, rows = self.cells.shape
        for cr in itertools.product(range(cols), range(rows)):
            self.plot_cell(cr, facecolor=(0.50, 0.50, 0.50, 1 / 16), ax=ax)

    def plot_cell_coordinates(self, ax: Axes, fontsize: int = 14) -> None:
        for coordinates in itertools.product(*(range(var) for var in self.cells.shape)):
            ax.text(
                *self._base_shape(*coordinates).to_plot_coordinates(),  # type: ignore
                s=f"{coordinates}",
                fontsize=fontsize,
                ha="center",
                va="center",
            )

    def plot_cell_concentration(
        self,
        ax: Axes,
        concentrations: NDArray,
        **kwargs: dict[str, Any],
    ) -> None:
        coordinates: tuple[int, int]
        concentrations = concentrations.reshape(self.gridsize)
        for coordinates in zip(*np.where(self.cells != 0)):  # type: ignore
            ax.text(
                *self._base_shape(*coordinates).to_plot_coordinates(),  # type: ignore
                s=f"{concentrations[coordinates]:.2g}",
                **{"ha": "center", "fontsize": 12} | kwargs,  # type: ignore
            )


@dataclass
class RodGrid(AbstractGrid):
    _base_shape: Type[Rod] = Rod
    _n_neighbors: int = 2

    def plot_axes(
        self,
        figsize: tuple[int, int] | None = None,
        ax: Axes | None = None,
    ) -> tuple[Figure, Axes]:
        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=figsize)
            ax = cast(Axes, ax)
        else:
            fig = cast(Figure, ax.get_figure())
        ax.set_aspect("equal")

        def pad(x: float) -> float:
            return x * 1.04

        ax.set_xlim(pad(-0.5), (self.cells.shape[0] + 0.5))
        ax.set_ylim(pad(-0.5), 0.5)
        ax.axis("off")
        return fig, ax

    def plot_grid(self, ax: Axes) -> None:
        cols = self.cells.shape[0]
        for col in range(cols):
            self.plot_cell((col,), facecolor=(0.50, 0.50, 0.50, 1 / 16), ax=ax)

    def plot_cell_coordinates(self, ax: Axes, fontsize: int = 14) -> None:
        for coordinates in range(self.cells.shape[0]):
            ax.text(
                *self._base_shape(coordinates).to_plot_coordinates(),  # type: ignore
                s=f"{coordinates}",
                fontsize=fontsize,
                ha="center",
                va="center",
            )


@dataclass
class TriGrid(AbstractGrid):
    _base_shape: Type[Triangle] = Triangle
    _n_neighbors: int = 3

    def plot_axes(
        self,
        figsize: tuple[int, int] | None = None,
        ax: Axes | None = None,
    ) -> tuple[Figure, Axes]:
        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=figsize)
            ax = cast(Axes, ax)
        else:
            fig = cast(Figure, ax.get_figure())
        ax.set_aspect("equal")

        c, r = self.cells.shape
        base_shape = cast(Triangle, self._base_shape(0, 0))
        x_offset = base_shape.x_offset
        y_offset = base_shape.y_offset
        x_lower = -x_offset
        x_upper = 0.5 * c
        y_lower = -y_offset
        y_upper = r * base_shape.h - y_offset
        ax.set_xlim(x_lower, x_upper)
        ax.set_ylim(y_lower, y_upper)
        ax.axis("off")
        return fig, ax


@dataclass
class SquareGrid(AbstractGrid):
    _base_shape: Type[Square] = Square
    _n_neighbors: int = 4

    def plot_axes(
        self,
        figsize: tuple[int, int] | None = None,
        ax: Axes | None = None,
    ) -> tuple[Figure, Axes]:
        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=figsize)
            ax = cast(Axes, ax)
        else:
            fig = cast(Figure, ax.get_figure())
        ax.set_aspect("equal")

        c, r = self.cells.shape
        x_offset = self._base_shape(0, 0).x_offset
        y_offset = self._base_shape(0, 0).y_offset

        x_lower = -x_offset - 0.1
        x_upper = c - x_offset + 0.1
        y_lower = -y_offset - 0.1
        y_upper = r - y_offset + 0.1
        ax.set_xlim(x_lower, x_upper)
        ax.set_ylim(y_lower, y_upper)
        ax.axis("off")
        return fig, ax


@dataclass
class HexGrid(AbstractGrid):
    _base_shape: Type[Oddq] = Oddq
    _n_neighbors: int = 6

    def plot_axes(
        self,
        figsize: tuple[int, int] | None = None,
        ax: Axes | None = None,
    ) -> tuple[Figure, Axes]:
        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=figsize)
            ax = cast(Axes, ax)
        else:
            fig = cast(Figure, ax.get_figure())
        ax.set_aspect("equal")

        c, r = self.cells.shape
        x_lower = -1
        x_upper = 1 + 3 / 2 * (c - 1)
        y_lower = -math.sqrt(3) * 0.5
        y_upper = math.sqrt(3) * r
        ax.set_xlim(x_lower, x_upper)
        ax.set_ylim(y_lower, y_upper)
        ax.axis("off")
        return fig, ax


@dataclass
class CubeGrid(AbstractGrid):
    _base_shape: Type[Cube3D] = Cube3D
    _n_neighbors: int = 6

    def plot_axes(
        self,
        figsize: tuple[int, int] | None = None,
        ax: Axes | None = None,
    ) -> tuple[Figure, Axes]:
        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=figsize, subplot_kw={"projection": "3d"})
            ax = cast(Axes, ax)
        else:
            fig = cast(Figure, ax.get_figure())

        a, b, c = self.cells.shape
        ax.set_xlim(0, a)
        ax.set_ylim(0, b)
        ax.set_zlim(0, c)  # type: ignore
        ax.axis("off")
        return fig, ax

    def plot_grid(
        self,
        ax: Axes,
        facecolor: str | tuple[float, float, float, float] = (0.50, 0.50, 0.50, 1 / 16),
        edgecolor: str | tuple[float, float, float, float] = (0, 0, 0, 1 / 16),
    ) -> None:
        x, y, z = self.cells.shape
        for xyz in itertools.product(range(x), range(y), range(z)):
            self.plot_cell(xyz, facecolor=facecolor, edgecolor=edgecolor, ax=ax)



from __future__ import annotations

from . import coordinates
from .grids import AbstractGrid, HexGrid, RodGrid, SquareGrid
from .models import MesophyllModel, StomataModel, VeinModel

__all__ = [
    "AbstractGrid",
    "HexGrid",
    "MesophyllModel",
    "RodGrid",
    "SquareGrid",
    "SquareGrid",
    "SquareGrid",
    "StomataModel",
    "VeinModel",
    "coordinates",
]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['dismo']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.8,<4.0',
 'numpy>=1.26,<2.0',
 'pandas>=2.1,<3.0',
 'pyarrow',
 'scipy>=1.12.0,<2.0.0',
 'typing-extensions>=4.9.0,<5.0.0']

setup_kwargs = {
    'name': 'dismo',
    'version': '1.0.45',
    'description': 'A subpackage of modelbase that enables investigation of PDE models',
    'long_description': '# DIscrete Spatial MOdels\n\n[![pipeline status](https://gitlab.com/qtb-hhu/dismo/badges/main/pipeline.svg)](https://gitlab.com/qtb-hhu/dismo/-/commits/main)\n[![coverage report](https://gitlab.com/qtb-hhu/dismo/badges/main/coverage.svg)](https://gitlab.com/qtb-hhu/dismo/-/commits/main)\n[![Documentation](https://img.shields.io/badge/Documentation-Gitlab-success)](https://qtb-hhu.gitlab.io/dismo/)\n[![PyPi](https://img.shields.io/pypi/v/dismo)](https://pypi.org/project/dismo/)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Checked with mypy](http://www.mypy-lang.org/static/mypy_badge.svg)](http://mypy-lang.org/)\n[![security: bandit](https://img.shields.io/badge/security-bandit-yellow.svg)](https://github.com/PyCQA/bandit)\n[![Downloads](https://pepy.tech/badge/dismo)](https://pepy.tech/project/dismo)\n\ndismo is a Python package for building and analysing discrete spatial models based on ordinary differential equations.\nIts primary purpose is to allow arbitrarily complex internal and transport processes to easily be mapped over multiple different regular grids.\nFor this it features one, two and three-dimensional layouts, with standard and non-standard (e.g. hexagonal or triangular) grids.\n\n\n## Installation\n\nIf you quickly want to test out dismo, or do not require assimulo support, install dismo via\n\n```bash\npip install dismo\n```\n\nTo enable assimulo support, the easiest way is to install it via mamba using the [mambaforge](https://github.com/conda-forge/miniforge#mambaforge) distribution\n\n```bash\nmamba create -n py311 python=3.11 assimulo\nmamba activate py311\npip install dismo\n```\n\n## License\n\n[GPL 3](https://gitlab.com/qtb-hhu/dismo/blob/main/LICENSE)\n\n## Documentation\n\nThe official documentation is hosted [here on gitlab](https://qtb-hhu.gitlab.io/dismo/).\n\n## Issues and support\n\nIf you experience issues using the software please contact us through our [issues](https://gitlab.com/qtb-hhu/dismo/issues) page.\n\n## Contributing to dismo\n\nAll contributions, bug reports, bug fixes, documentation improvements, enhancements and ideas are welcome.\nSee our [contribution guide](https://gitlab.com/qtb-hhu/dismo/blob/main/CONTRIBUTING.md) for more information.\n\n## How to cite\n\ndismo is currently in the publication process. You will find citing information here as soon as possible.\n\n<!-- If you use this software in your scientific work, please cite [this article](https://rdcu.be/ckOSa):\n\nvan Aalst, M., Ebenhöh, O. & Matuszyńska, A. Constructing and analysing dynamic models with dismo v1.2.3: a software update. BMC Bioinformatics 22, 203 (2021)\n\n- [doi](https://doi.org/10.1186/s12859-021-04122-7)\n- [bibtex file](https://gitlab.com/qtb-hhu/dismo/blob/main/citation.bibtex) -->\n',
    'author': 'Marvin van Aalst',
    'author_email': 'marvin.vanaalst@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.11,<3.13',
}


setup(**setup_kwargs)
